/*
  See the file "pnpoly.c" for copyright and license information.
*/

/*
  Returns 1 if the point (x,y) is inside the polygon defined by the set of points
  (xp,yp) of length "npol".
 */
int point_in_poly(double* xp, double* yp, int npol, double x, double y);
