/*
# This file is part of the Astrometry.net suite.
# Licensed under a 3-clause BSD style license - see LICENSE
*/

#ifndef TABSORT_H
#define TABSORT_H

int tabsort(const char* infn, const char* outfn, const char* colname,
            int descending);

#endif
